// Usage:
// $ gcc -Wall -o domplay domplay.c -lSDL -I/usr/include/SDL
// No color support.
// No sane attempt was made at audio sync. Frames last roughly as long as they should,
// typically a bit longer, so expect some audio underruns (butthe video should mostly
// stay in sync).
// Without audio:
// $ ./domplay BAFULL.XDV > /dev/null
// With audio (pipe to something receiving 22050Hz 8-bit unsigned audio)
// $ ./domplay BAFULL.XDV | play -r22050 -c1 -u -1 -traw -
// Too lazy to use SDL for audio because I'd have to buffer video and audio myself.

#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <stdint.h>
#include <malloc.h>
#include <unistd.h>

#include "SDL/SDL.h"

int main(int argc, char **argv)
{
    FILE *fp = fopen(argv[1], "rb");

    fseek(fp, 0, SEEK_END);
    size_t size = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    uint8_t *data = memalign(0x200, size);

    fread(data, size, 1, fp);
    assert(!memcmp("XDCV", data, 4));

    uint8_t *p = data + 0x200;
    uint8_t *end = data + size;

    const uint8_t hdr[] = {0x1e, 0x0e, 0x1f};


    uint8_t vram[0x10000];

    SDL_Surface* screen = NULL;
    SDL_Init(SDL_INIT_EVERYTHING);
    screen = SDL_SetVideoMode(640, 400, 32, SDL_SWSURFACE);
    uint32_t *pixels = (uint32_t *)screen->pixels;

    int nframes = data[4] | (data[5] << 8);
    int audio_block = data[8] | (data[9] << 8);
    int sample_rate = data[10] | (data[11] << 8);
    fprintf(stderr, "number of frames: %d\n", nframes);
    fprintf(stderr, "audio: 0x%x samples per frame, %dHz\n", audio_block, sample_rate);
    int frame_time = 1000 * audio_block / sample_rate;
    fprintf(stderr, "frame time: %dms\n", frame_time);

    uint8_t *fsizes = end - nframes;

    int frame = 0;
    while (frame < nframes) {
        fprintf(stderr,"Frame %d at %08lx\n", frame, p-data);
        assert(!memcmp(p, hdr, sizeof(hdr)));
        uint16_t memoff = (p[5]<<8) | p[4];
        uint8_t *dp = p + memoff;
        uint8_t *ip = p + 0x0c;
        uint16_t di = 0;
        uint8_t val = 0;
        uint16_t cnt;
        while (*ip != 0x1f) {
            switch (*ip) {
                case 0xbf:
                    // mov di
                    di = (ip[2] << 8) | ip[1];
                    ip += 3;
                    break;
                case 0xb1:
                    // mov cl
                    cnt = ip[1];
                    ip += 2;
                    break;
                case 0xb0:
                    // mov al
                    val = ip[1];
                    ip += 2;
                    break;
                case 0xb9:
                    // mov cx
                    cnt = (ip[2] << 8) | ip[1];
                    ip += 3;
                    break;
                case 0xf3:
                    switch (ip[1]) {
                        case 0xa4:
                            // rep movsb
                            memcpy(vram + di, dp, cnt);
                            dp += cnt;
                            di += cnt;
                            break;
                        case 0xaa:
                            // rep stosb
                            memset(vram + di, val, cnt);
                            di += cnt;
                            break;
                        default:
                            fprintf(stderr, "Unhandled opcode: %02x at %08lx\n", *ip, ip-data);
                            return 1;
                    }
                    ip += 2;
                    break;
                case 0xa5:
                    // movsw
                    memcpy(vram + di, dp, 2);
                    dp += 2;
                    di += 2;
                    ip++;
                    break;
                case 0xa4:
                    // movsb
                    vram[di] = *dp;
                    dp += 1;
                    di += 1;
                    ip++;
                    break;
                default:
                    fprintf(stderr, "Unhandled opcode: %02x at %08lx\n", *ip, ip-data);
                    return 1;
            }
        }
        assert (ip[1] == 0xcb);
        //fprintf(stderr, "DP at %08lx\n", dp-data);
        //dp += audio_block;
        //p = (uint8_t *)((((intptr_t)dp) + 0x1ff) & ~0x1ff);
        p += fsizes[frame] * 0x200;
        write(1, p-audio_block, audio_block);

        int x, y, i, b, s;
        s = screen->w;
        for (y=0; y<200; y+=2) {
            for (i=0; i<2; i++) {
                for (x=0; x<640; x+=8) {
                    for (b=0; b<8; b++) {
                        pixels[2*(y+i)*s + x + b] = (vram[y*40 + i*8192 + x/8] & (0x80>>b)) ? 0xffffffff : 0xff000000;
                        pixels[(2*(y+i)+1)*s + x + b] = (vram[y*40 + i*8192 + x/8] & (0x80>>b)) ? 0xffffffff : 0xff000000;
                    }
                }
            }
        }

        SDL_Flip(screen);
        SDL_Delay(frame_time);
        frame++;
    }

    SDL_Quit();
    return 0;
}